
import uuid
import json
import time
from queue import Queue

from . import utils, mqtt_client
from .webapp.socketing import update_online_dot, update_late_badge


class RemoteDevice(object):
    """
    Implementation of a serial device remotely connected to an MQTT broker
    """

    def __init__(self, topic_name, mode, mute=False, accepts_incoming=True, name="Remote Device", on_time_max=0,
                 translation=[False, None, None, 0]):
        self.sku = id(self)
        if name[0] == "$":
            raise ValueError("Device name cannot start with '$'.")
        self.name = name
        self.mute = mute
        self.accepts_incoming = accepts_incoming
        self.type = "mqtt"
        self.type_color = "cyan"
        parsed_topic = topic_name.split("/")
        self.remote_client = parsed_topic[0]

        # if a device name was not given, subscribe to all devices on remote
        if len(parsed_topic) > 1:
            self.remote_device_name = parsed_topic[1]
        else:
            self.remote_device_name = "+"
            # cannot publish to device if no device name given
            self.accepts_incoming = False

        self.topic_name = f"{self.remote_client}/{self.remote_device_name}"
        self.mode = mode
        self.message_queue = Queue()
        self.late_message_list= []
        self._last_message = ''
        self.online = False
        self.listening = True
        self.remove_me = False
        self.translation = translation
        self.on_time_max = int(on_time_max)
        mqtt_client.client.message_callback_add(self.topic_name+"/from_device", self._new_message)
        mqtt_client.add_subscription(self.topic_name+"/#")
        mqtt_client.subscribe_status(self.remote_client, self._update_status)
        self.max_transit_time = 0
        self.min_transit_time = 9999
        self.average_transit_time = 0
        self.received_messages_count = 0

    def _update_status(self, status):
        if status == b"online":
            self.online = True
        else:
            self.online = False
        update_online_dot(self.sku, self.online)
    
    def _update_transit_stats(self, transit_time):
        self.min_transit_time = min(self.min_transit_time, transit_time)
        self.max_transit_time = max(self.max_transit_time, transit_time)
        sum_recalc = self.average_transit_time * self.received_messages_count
        self.received_messages_count += 1
        self.average_transit_time = round((sum_recalc + transit_time)/self.received_messages_count, 2)

    def _new_message(self, client, userdata, message):
        if self.listening:
            try:
                payload = json.loads(message.payload.decode())
                now_corrected = time.time() + mqtt_client.time_offset
                message_transit_time = round(now_corrected - payload["timestamp"], 1)
                self._update_transit_stats(message_transit_time)
                message_data = payload["message"]
                stripped_message = " ".join(message_data.split())
                utils.logger.info(f"Message {stripped_message} transit time: {message_transit_time}.")
                if self.on_time_max == 0 or message_transit_time < self.on_time_max:
                    self.message_queue.put(message_data.encode())
                else:
                    self.late_message_list.append(message_data.encode())
                    utils.print_to_ui(f"Received message from {self.name} "
                                    f"after max on-time window of {self.on_time_max}s ({message_transit_time}s). "
                                    f"Adding to late message list.")
                    update_late_badge(self.sku, len(self.late_message_list))
            except Exception as e:
                utils.print_to_ui(f"Error receiving remote message: {e}")

    def write(self, message):
        if self.accepts_incoming and self.remote_device_name != "+":
            mqtt_client.client.publish(self.topic_name+"/from_remote", message, qos=1)
            return True
        return False

    def accept_late_messages(self, message_index_list=None):
        if message_index_list is not None:
            temp_list = []
            message_index_list.sort(reverse=True)
            for index in message_index_list:
                temp_list.append(self.late_message_list.pop(index))
            while len(temp_list) > 0:
                self.message_queue.put(temp_list.pop())
        else:
            for message in self.late_message_list:
                self.message_queue.put(message)
                self.late_message_list.remove(message)
        update_late_badge(self.sku, len(self.late_message_list))

    def kill_listen_stream(self):
        self.listening = False


import datetime
import os
import uuid
import socket
import re
import sys
import glob
import serial
import logging
import logging.handlers

from colorama import Fore, Style
from colorama import init as colorama_init

from .settings import USER_SPECIFIED_DEFAULT_CONFIG_FILE, USER_SPECIFIED_LOG_FILE
from .webapp import socketing
from .cli import interface

colorama_init()
ui = "webapp"

def set_ui(user_int):
    global ui
    ui = user_int


def get_default_config_file():
    if USER_SPECIFIED_DEFAULT_CONFIG_FILE is None:
        home_path = os.path.expanduser('~')
        # check if we're on linux running as superuser, then choose more appropriate directory
        if home_path == "/root":
            home_path = "/etc"
        else:
            home_path = os.path.join(home_path, '.config')
        file_name = "config.ini"
        return os.path.join(home_path, 'ullr', file_name)
    else:
        return USER_SPECIFIED_DEFAULT_CONFIG_FILE


def get_log_file():
    if USER_SPECIFIED_LOG_FILE is None:
        home_path = os.path.expanduser('~')
        # check if we're on linux running as superuser, then choose more appropriate directory
        if home_path == "/root":
            home_path = "/var/log"
        else:
            home_path = os.path.join(home_path, '.log')
        file_name = "ullr.log"
        path = os.path.join(home_path, 'ullr', file_name)
    else:
        path = USER_SPECIFIED_DEFAULT_CONFIG_FILE
    os.makedirs(os.path.dirname(path), exist_ok=True)
    return path


def setup_logger():
    logger = logging.getLogger("")
    logger.setLevel(logging.DEBUG)
    logging.getLogger("requests").setLevel(logging.WARNING)
    logging.getLogger("urllib3").setLevel(logging.WARNING)
    handler = logging.handlers.RotatingFileHandler(
        get_log_file(), maxBytes=(1024 * 100), backupCount=5
    )
    formatter = logging.Formatter(
        "%(asctime)s - %(name)s - %(levelname)s - %(message)s")
    handler.setFormatter(formatter)
    logger.addHandler(handler)
    return logger


logger = setup_logger()

sys_stamp = "[ " + Fore.LIGHTBLACK_EX + "sys" + Style.RESET_ALL + " ] "


def timestamp():
    return "[" + Fore.LIGHTBLACK_EX + datetime.datetime.now().strftime("%Y-%m-%d %H:%M:%S") + Style.RESET_ALL + "] "


def internet_connection(host="8.8.8.8", port=53, timeout=3):
    """
        Host: 8.8.8.8 (google-public-dns-a.google.com)
        OpenPort: 53/tcp
        Service: domain (DNS/TCP)
        """
    try:
        socket.setdefaulttimeout(timeout)
        socket.socket(socket.AF_INET, socket.SOCK_STREAM).connect((host, port))
        return True
    except socket.error as ex:
        #  print(ex)
        return False


def get_ip():
    s = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
    try:
        # doesn't even have to be reachable
        s.connect(('8.8.8.8', 1))
        IP = s.getsockname()[0]
    except Exception:
        IP = '127.0.0.1'
    finally:
        s.close()
    return IP


def get_available_com_ports():
    """ Lists serial port names

        :raises EnvironmentError:
            On unsupported or unknown platforms
        :returns:
            A list of the serial ports available on the system
    """
    if sys.platform.startswith('win'):
        ports = ['COM%s' % (i + 1) for i in range(100)]
    elif sys.platform.startswith('linux') or sys.platform.startswith('cygwin'):
        # this excludes your current terminal "/dev/tty"
        ports = glob.glob('/dev/tty[A-Za-z]*')
        for item in glob.glob('/dev/serial[0-9]*'):
            ports.append(item)
    elif sys.platform.startswith('darwin'):
        ports = glob.glob('/dev/tty.*')
    else:
        raise EnvironmentError('Unsupported platform')

    result = []
    for port in ports:
        try:
            s = serial.Serial(port)
            s.close()
            result.append(port)
        except (OSError, serial.SerialException):
            pass
    return result


def print_to_ui(message, endline="\n", sys=False):
    ansi_escape = re.compile(r'\x1B(?:[@-Z\\-_]|\[[0-?]*[ -/]*[@-~])')
    no_colors = ansi_escape.sub('', message)
    logger.info(no_colors)
    if sys:
        message = sys_stamp + message
    else:
        message = timestamp() + message
    if ui == "webapp":
        message = ansi_escape.sub('', message)
        socketing.print_to_web_console(message, endline=endline)
    elif ui == "cli":
        interface.s_print(message, end=endline)



# -*- coding: utf-8 -*-
# http://multivax.com/last_question.html

"""Get the MAC address of remote hosts or network interfaces.
It provides a platform-independent interface to get the MAC addresses of:
* System network interfaces (by interface name)
* Remote hosts on the local network (by IPv4/IPv6 address or hostname)
It provides one function: `get_mac_address()`
Examples:
    from getmac import get_mac_address
    eth_mac = get_mac_address(interface="eth0")
    win_mac = get_mac_address(interface="Ethernet 3")
    ip_mac = get_mac_address(ip="192.168.0.1")
    ip6_mac = get_mac_address(ip6="::1")
    host_mac = get_mac_address(hostname="localhost")
    updated_mac = get_mac_address(ip="10.0.0.1", network_request=True)
"""

import ctypes
import logging
import os
import platform
import re
import shlex
import socket
import struct
import sys
import traceback
from subprocess import check_output

try:  # Python 3
    from subprocess import DEVNULL  # type: ignore
except ImportError:  # Python 2
    DEVNULL = open(os.devnull, "wb")  # type: ignore

# Configure logging
log = logging.getLogger("getmac")
log.addHandler(logging.NullHandler())

__version__ = "0.8.2"
PY2 = sys.version_info[0] == 2

# Configurable settings
DEBUG = 0
PORT = 55555

# Platform identifiers
_SYST = platform.system()
if _SYST == "Java":
    try:
        import java.lang

        _SYST = str(java.lang.System.getProperty("os.name"))
    except ImportError:
        log.critical("Can't determine OS: couldn't import java.lang on Jython")
WINDOWS = _SYST == "Windows"
DARWIN = _SYST == "Darwin"
OPENBSD = _SYST == "OpenBSD"
FREEBSD = _SYST == "FreeBSD"
BSD = OPENBSD or FREEBSD  # Not including Darwin for now
WSL = False  # Windows Subsystem for Linux (WSL)
LINUX = False
if _SYST == "Linux":
    if "Microsoft" in platform.version():
        WSL = True
    else:
        LINUX = True

PATH = os.environ.get("PATH", os.defpath).split(os.pathsep)
if not WINDOWS:
    PATH.extend(("/sbin", "/usr/sbin"))

# Use a copy of the environment so we don't
# modify the process's current environment.
ENV = dict(os.environ)
ENV["LC_ALL"] = "C"  # Ensure ASCII output so we parse correctly

ARP_PATH = os.environ.get("ARP_PATH", "/proc/net/arp")

# Constants
IP4 = 0
IP6 = 1
INTERFACE = 2
HOSTNAME = 3

MAC_RE_COLON = r"([0-9a-fA-F]{2}(?::[0-9a-fA-F]{2}){5})"
MAC_RE_DASH = r"([0-9a-fA-F]{2}(?:-[0-9a-fA-F]{2}){5})"
MAC_RE_DARWIN = r"([0-9a-fA-F]{1,2}(?::[0-9a-fA-F]{1,2}){5})"

# Used for mypy (a data type analysis tool)
# If you're copying the code, this section can be safely removed
try:
    from typing import TYPE_CHECKING

    if TYPE_CHECKING:
        from typing import Optional
except ImportError:
    pass

# Ensure we only log the Python 2 warning once
WARNED_PY2 = False


def get_mac_address(
    interface=None, ip=None, ip6=None, hostname=None, network_request=True
):
    # type: (Optional[str], Optional[str], Optional[str], Optional[str], bool) -> Optional[str]
    """Get a Unicast IEEE 802 MAC-48 address from a local interface or remote host.
    You must only use one of the first four arguments. If none of the arguments
    are selected, the default network interface for the system will be used.
    Exceptions will be handled silently and returned as a None.
    For the time being, it assumes you are using Ethernet.
    NOTES:
    * You MUST provide str-typed arguments, REGARDLESS of Python version.
    * localhost/127.0.0.1 will always return '00:00:00:00:00:00'
    Args:
        interface (str): Name of a local network interface (e.g "Ethernet 3", "eth0", "ens32")
        ip (str): Canonical dotted decimal IPv4 address of a remote host (e.g 192.168.0.1)
        ip6 (str): Canonical shortened IPv6 address of a remote host (e.g ff02::1:ffe7:7f19)
        hostname (str): DNS hostname of a remote host (e.g "router1.mycorp.com", "localhost")
        network_request (bool): Send a UDP packet to a remote host to populate
        the ARP/NDP tables for IPv4/IPv6. The port this packet is sent to can
        be configured using the module variable `getmac.PORT`.
    Returns:
        Lowercase colon-separated MAC address, or None if one could not be
        found or there was an error.
    """
    if PY2:
        global WARNED_PY2
        if not WARNED_PY2:
            log.warning(
                "Python 2 compatibility will be dropped in getmac 1.0.0. If you are "
                'stuck on Python 2, consider loosely pinning the version e.g. "getmac<1".'
            )
            WARNED_PY2 = True

    if (hostname and hostname == "localhost") or (ip and ip == "127.0.0.1"):
        return "00:00:00:00:00:00"

    # Resolve hostname to an IP address
    if hostname:
        ip = socket.gethostbyname(hostname)

    # Populate the ARP table by sending a empty UDP packet to a high port
    if network_request and (ip or ip6):
        if ip:
            s = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
        else:
            s = socket.socket(socket.AF_INET6, socket.SOCK_DGRAM)
        try:
            if ip:
                s.sendto(b"", (ip, PORT))
            else:
                s.sendto(b"", (ip6, PORT))
        except Exception:
            log.error("Failed to send ARP table population packet")
            if DEBUG:
                log.debug(traceback.format_exc())
        finally:
            s.close()

    # Setup the address hunt based on the arguments specified
    if ip6:
        if not socket.has_ipv6:
            log.error(
                "Cannot get the MAC address of a IPv6 host: "
                "IPv6 is not supported on this system"
            )
            return None
        elif ":" not in ip6:
            log.error("Invalid IPv6 address: %s", ip6)
            return None
        to_find = ip6
        typ = IP6
    elif ip:
        to_find = ip
        typ = IP4
    else:  # Default to searching for interface
        typ = INTERFACE
        if interface:
            to_find = interface
        else:
            # Default to finding MAC of the interface with the default route
            if WINDOWS and network_request:
                to_find = _fetch_ip_using_dns()
                typ = IP4
            elif WINDOWS:
                to_find = "Ethernet"
            elif BSD:
                if OPENBSD:
                    to_find = _get_default_iface_openbsd()  # type: ignore
                else:
                    to_find = _get_default_iface_freebsd()  # type: ignore
                if not to_find:
                    to_find = "em0"
            else:
                to_find = _hunt_linux_default_iface()  # type: ignore
                if not to_find:
                    to_find = "en0"

    mac = _hunt_for_mac(to_find, typ, network_request)
    log.debug("Raw MAC found: %s", mac)

    # Check and format the result to be lowercase, colon-separated
    if mac is not None:
        mac = str(mac)
        if not PY2:  # Strip bytestring conversion artifacts
            for garbage_string in ["b'", "'", "\\n", "\\r"]:
                mac = mac.replace(garbage_string, "")
        mac = mac.strip().lower().replace(" ", "").replace("-", ":")

        # Fix cases where there are no colons
        if ":" not in mac and len(mac) == 12:
            log.debug("Adding colons to MAC %s", mac)
            mac = ":".join(mac[i : i + 2] for i in range(0, len(mac), 2))

        # Pad single-character octets with a leading zero (e.g Darwin's ARP output)
        elif len(mac) < 17:
            log.debug(
                "Length of MAC %s is %d, padding single-character " "octets with zeros",
                mac,
                len(mac),
            )
            parts = mac.split(":")
            new_mac = []
            for part in parts:
                if len(part) == 1:
                    new_mac.append("0" + part)
                else:
                    new_mac.append(part)
            mac = ":".join(new_mac)

        # MAC address should ALWAYS be 17 characters before being returned
        if len(mac) != 17:
            log.warning("MAC address %s is not 17 characters long!", mac)
            mac = None
        elif mac.count(":") != 5:
            log.warning("MAC address %s is missing ':' characters", mac)
            mac = None
    return mac


def _search(regex, text, group_index=0):
    # type: (str, str, int) -> Optional[str]
    match = re.search(regex, text)
    if match:
        return match.groups()[group_index]
    return None


def _popen(command, args):
    # type: (str, str) -> str
    for directory in PATH:
        executable = os.path.join(directory, command)
        if (
            os.path.exists(executable)
            and os.access(executable, os.F_OK | os.X_OK)
            and not os.path.isdir(executable)
        ):
            break
    else:
        executable = command
    if DEBUG >= 3:
        log.debug("Running: '%s %s'", executable, args)
    return _call_proc(executable, args)


def _call_proc(executable, args):
    # type: (str, str) -> str
    if WINDOWS:
        cmd = executable + " " + args  # type: ignore
    else:
        cmd = [executable] + shlex.split(args)  # type: ignore
    output = check_output(cmd, stderr=DEVNULL, env=ENV)
    if DEBUG >= 4:
        log.debug("Output from '%s' command: %s", executable, str(output))
    if not PY2 and isinstance(output, bytes):
        return str(output, "utf-8")
    else:
        return str(output)


def _windows_ctypes_host(host):
    # type: (str) -> Optional[str]
    if not PY2:  # Convert to bytes on Python 3+ (Fixes GitHub issue #7)
        host = host.encode()  # type: ignore
    try:
        inetaddr = ctypes.windll.wsock32.inet_addr(host)  # type: ignore
        if inetaddr in (0, -1):
            raise Exception
    except Exception:
        hostip = socket.gethostbyname(host)
        inetaddr = ctypes.windll.wsock32.inet_addr(hostip)  # type: ignore

    buffer = ctypes.c_buffer(6)
    addlen = ctypes.c_ulong(ctypes.sizeof(buffer))

    send_arp = ctypes.windll.Iphlpapi.SendARP  # type: ignore
    if send_arp(inetaddr, 0, ctypes.byref(buffer), ctypes.byref(addlen)) != 0:
        return None

    # Convert binary data into a string.
    macaddr = ""
    for intval in struct.unpack("BBBBBB", buffer):  # type: ignore
        if intval > 15:
            replacestr = "0x"
        else:
            replacestr = "x"
        macaddr = "".join([macaddr, hex(intval).replace(replacestr, "")])
    return macaddr


def _fcntl_iface(iface):
    # type: (str) -> str
    import fcntl

    if not PY2:
        iface = iface.encode()  # type: ignore
    s = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
    # 0x8927 = SIOCGIFADDR
    info = fcntl.ioctl(s.fileno(), 0x8927, struct.pack("256s", iface[:15]))
    if PY2:
        return ":".join(["%02x" % ord(char) for char in info[18:24]])
    else:
        return ":".join(["%02x" % ord(chr(char)) for char in info[18:24]])


def _uuid_ip(ip):
    # type: (str) -> Optional[str]
    from uuid import _arp_getnode  # type: ignore

    backup = socket.gethostbyname
    try:
        socket.gethostbyname = lambda x: ip
        mac1 = _arp_getnode()
        if mac1 is not None:
            mac1 = _uuid_convert(mac1)
            mac2 = _arp_getnode()
            mac2 = _uuid_convert(mac2)
            if mac1 == mac2:
                return mac1
    except Exception:
        raise
    finally:
        socket.gethostbyname = backup
    return None


def _uuid_lanscan_iface(iface):
    # type: (str) -> Optional[str]
    from uuid import _find_mac  # type: ignore

    if not PY2:
        iface = bytes(iface, "utf-8")  # type: ignore
    mac = _find_mac("lanscan", "-ai", [iface], lambda i: 0)
    if mac:
        return _uuid_convert(mac)
    return None


def _uuid_convert(mac):
    # type: (int) -> str
    return ":".join(("%012X" % mac)[i : i + 2] for i in range(0, 12, 2))


def _read_sys_iface_file(iface):
    # type: (str) -> Optional[str]
    data = _read_file("/sys/class/net/" + iface + "/address")
    # Sometimes this can be empty or a single newline character
    return None if data is not None and len(data) < 17 else data


def _read_arp_file(host):
    # type: (str) -> Optional[str]
    data = _read_file(ARP_PATH)
    if data is not None and len(data) > 1:
        # Need a space, otherwise a search for 192.168.16.2
        # will match 192.168.16.254 if it comes first!
        return _search(re.escape(host) + r" .+" + MAC_RE_COLON, data)
    return None


def _arping_habets(host):
    # type: (str) -> Optional[str]
    """Parse https://github.com/ThomasHabets/arping output."""
    return _search(r"^%s$" % MAC_RE_COLON, _popen("arping", "-r -C 1 -c 1 %s" % host),)


def _arping_iputils(host):
    # type: (str) -> Optional[str]
    """Parse iputils arping output."""
    return _search(
        r" from %s \[(%s)\]" % (re.escape(host), MAC_RE_COLON),
        _popen("arping", "-f -c 1 %s" % host),
    )


def _read_file(filepath):
    # type: (str) -> Optional[str]
    try:
        with open(filepath) as f:
            return f.read()
    except (OSError, IOError):  # This is IOError on Python 2.7
        log.debug("Could not find file: '%s'", filepath)
        return None


def _hunt_for_mac(to_find, type_of_thing, net_ok=True):
    # type: (Optional[str], int, bool) -> Optional[str]
    """Tries a variety of methods to get a MAC address.
    Format of method lists:
    Tuple:  (regex, regex index, command, command args)
            Command args is a list of strings to attempt to use as arguments
    lambda: Function to call
    """
    if to_find is None:
        log.warning("_hunt_for_mac() failed: to_find is None")
        return None
    if not PY2 and isinstance(to_find, bytes):
        to_find = str(to_find, "utf-8")

    if WINDOWS and type_of_thing == INTERFACE:
        methods = [
            # getmac - Connection Name
            (
                r"\r\n" + to_find + r".*" + MAC_RE_DASH + r".*\r\n",
                0,
                "getmac.exe",
                ["/NH /V"],
            ),
            # ipconfig
            (
                to_find
                + r"(?:\n?[^\n]*){1,8}Physical Address[ .:]+"
                + MAC_RE_DASH
                + r"\r\n",
                0,
                "ipconfig.exe",
                ["/all"],
            ),
            # getmac - Network Adapter (the human-readable name)
            (
                r"\r\n.*" + to_find + r".*" + MAC_RE_DASH + r".*\r\n",
                0,
                "getmac.exe",
                ["/NH /V"],
            ),
            # wmic - WMI command line utility
            lambda x: _popen(
                "wmic.exe",
                "nic where \"NetConnectionID = '%s'\" get " "MACAddress /value" % x,
            )
            .strip()
            .partition("=")[2],
        ]
    elif (WINDOWS or WSL) and type_of_thing in [IP4, IP6, HOSTNAME]:
        methods = [
            # arp -a - Parsing result with a regex
            (MAC_RE_DASH, 0, "arp.exe", ["-a %s" % to_find])
        ]

        # Add methods that make network requests
        # Insert it *after* arp.exe since that's probably faster.
        if net_ok and type_of_thing != IP6 and not WSL:
            methods.insert(1, _windows_ctypes_host)
    elif (DARWIN or FREEBSD) and type_of_thing == INTERFACE:
        methods = [
            (r"ether " + MAC_RE_COLON, 0, "ifconfig", [to_find]),
            # Alternative match for ifconfig if it fails
            (to_find + r".*ether " + MAC_RE_COLON, 0, "ifconfig", [""]),
            (MAC_RE_COLON, 0, "networksetup", ["-getmacaddress %s" % to_find]),
        ]
    elif FREEBSD and type_of_thing in [IP4, IP6, HOSTNAME]:
        methods = [
            (
                r"\(" + re.escape(to_find) + r"\)\s+at\s+" + MAC_RE_COLON,
                0,
                "arp",
                [to_find],
            )
        ]
    elif OPENBSD and type_of_thing == INTERFACE:
        methods = [(r"lladdr " + MAC_RE_COLON, 0, "ifconfig", [to_find])]
    elif OPENBSD and type_of_thing in [IP4, IP6, HOSTNAME]:
        methods = [(re.escape(to_find) + r"[ ]+" + MAC_RE_COLON, 0, "arp", ["-an"])]
    elif type_of_thing == INTERFACE:
        methods = [
            _read_sys_iface_file,
            _fcntl_iface,
            # Fast modern Ubuntu ifconfig
            (r"ether " + MAC_RE_COLON, 0, "ifconfig", [to_find]),
            # Fast ifconfig
            (r"HWaddr " + MAC_RE_COLON, 0, "ifconfig", [to_find]),
            # ip link (Don't use 'list' due to SELinux [Android 24+])
            (
                to_find + r".*\n.*link/ether " + MAC_RE_COLON,
                0,
                "ip",
                ["link %s" % to_find, "link"],
            ),
            # netstat
            (to_find + r".*HWaddr " + MAC_RE_COLON, 0, "netstat", ["-iae"]),
            # More variations of ifconfig
            (to_find + r".*ether " + MAC_RE_COLON, 0, "ifconfig", [""]),
            (to_find + r".*HWaddr " + MAC_RE_COLON, 0, "ifconfig", ["", "-a", "-v"]),
            # Tru64 ('-av')
            (to_find + r".*Ether " + MAC_RE_COLON, 0, "ifconfig", ["-av"]),
            _uuid_lanscan_iface,
        ]
    elif type_of_thing in [IP4, IP6, HOSTNAME]:
        esc = re.escape(to_find)
        methods = [
            _read_arp_file,
            lambda x: _popen("ip", "neighbor show %s" % x)
            .partition(x)[2]
            .partition("lladdr")[2]
            .strip()
            .split()[0],
            (
                r"\(" + esc + r"\)\s+at\s+" + MAC_RE_COLON,
                0,
                "arp",
                [to_find, "-an", "-an %s" % to_find],
            ),
            # Darwin oddness
            (
                r"\(" + esc + r"\)\s+at\s+" + MAC_RE_DARWIN,
                0,
                "arp",
                [to_find, "-a", "-a %s" % to_find],
            ),
            _uuid_ip,
        ]
        # Add methods that make network requests
        if net_ok and type_of_thing != IP6:
            methods.extend((_arping_iputils, _arping_habets))
    else:
        log.critical("Reached end of _hunt_for_mac() if-else chain!")
        return None
    return _try_methods(methods, to_find)


def _try_methods(methods, to_find=None):
    # type: (list, Optional[str]) -> Optional[str]
    """Runs the methods specified by _hunt_for_mac().
    We try every method and see if it returned a MAC address. If it returns
    None or raises an exception, we continue and try the next method.
    """
    found = None
    for m in methods:
        try:
            if isinstance(m, tuple):
                for arg in m[3]:  # list(str)
                    if DEBUG:
                        log.debug("Trying: '%s %s'", m[2], arg)
                    # Arguments: (regex, _popen(command, arg), regex index)
                    found = _search(m[0], _popen(m[2], arg), m[1])
                    if DEBUG:
                        log.debug("Result: %s\n", found)
                    if found:  # Skip remaining args AND remaining methods
                        break
            elif callable(m):
                if DEBUG:
                    log.debug("Trying: '%s' (to_find: '%s')", m.__name__, str(to_find))
                if to_find is not None:
                    found = m(to_find)
                else:
                    found = m()
                if DEBUG:
                    log.debug("Result: %s\n", found)
            else:
                log.critical("Invalid type '%s' for method '%s'", type(m), str(m))
        except Exception as ex:
            if DEBUG:
                log.debug("Exception: %s", str(ex))
            if DEBUG >= 2:
                log.debug(traceback.format_exc())
            continue
        if found:  # Skip remaining methods
            break
    return found


def _get_default_iface_linux():
    # type: () -> Optional[str]
    """Get the default interface by reading /proc/net/route.
    This is the same source as the `route` command, however it's much
    faster to read this file than to call `route`. If it fails for whatever
    reason, we can fall back on the system commands (e.g for a platform
    that has a route command, but maybe doesn't use /proc?).
    """
    data = _read_file("/proc/net/route")
    if data is not None and len(data) > 1:
        for line in data.split("\n")[1:-1]:
            iface_name, dest = line.split("\t")[:2]
            if dest == "00000000":
                return iface_name
    return None


def _hunt_linux_default_iface():
    # type: () -> Optional[str]
    # NOTE: for now, we check the default interface for WSL using the
    # same methods as POSIX, since those parts of the net stack work fine.
    methods = [
        _get_default_iface_linux,
        lambda: _popen("route", "-n")
        .partition("0.0.0.0")[2]
        .partition("\n")[0]
        .split()[-1],
        lambda: _popen("ip", "route list 0/0")
        .partition("dev")[2]
        .partition("proto")[0]
        .strip(),
    ]
    return _try_methods(methods)


def _get_default_iface_openbsd():
    # type: () -> Optional[str]
    methods = [
        lambda: _popen("route", "-nq show -inet -gateway -priority 1")
        .partition("127.0.0.1")[0]
        .strip()
        .rpartition(" ")[2]
    ]
    return _try_methods(methods)


def _get_default_iface_freebsd():
    # type: () -> Optional[str]
    methods = [(r"default[ ]+\S+[ ]+\S+[ ]+(\S+)\n", 0, "netstat", ["-r"])]
    return _try_methods(methods)


def _fetch_ip_using_dns():
    # type: () -> str
    """Determines the IP address of the default network interface.
    Sends a UDP packet to Cloudflare's DNS (1.1.1.1), which should go through
    the default interface. This populates the source address of the socket,
    which we then inspect and return.
    """
    s = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
    s.connect(("1.1.1.1", 53))
    ip = s.getsockname()[0]
    s.close()  # NOTE: sockets don't have context manager in 2.7 :(
    return ip



from threading import Lock, Thread
import time

import paho.mqtt.client as mqtt
import ntplib
import certifi

from .utils import logger, print_to_ui, internet_connection
from .webapp import socketing

mqtt_broker_url, mqtt_broker_user, mqtt_broker_pw, CLIENT_ID = "", "", "", ""
mqtt_broker_port = 0

client = None

connected = False
time_offset = 0

status_functions = {}
subscriptions = []
ntp_client = ntplib.NTPClient()
publish_lock = Lock()


def update_time_offset():
    while not internet_connection():
        time.sleep(2)
    global time_offset
    try:
        resp = ntp_client.request('pool.ntp.org')
        time_offset = resp.offset
        logger.info(f"Time offset set to {time_offset}.")
        return True
    except:
        logger.warn("Unable to update time offset.")
        return update_time_offset()


def safe_publish(*args, **kwargs):
    with publish_lock:
        client.publish(*args, **kwargs)


def on_connect(client, userdata, flags, rc):
    global connected

    t = Thread(target=update_time_offset)
    t.daemon = True
    # t.start()  # removed due to inaccuracies running on raspberry pi at boot
    if rc == 0:
        connected = True
        safe_publish(CLIENT_ID+"/status", "online", qos=1, retain=True)
        for topic in subscriptions:
            client.subscribe(topic, qos=1)
        print_to_ui("Connected to MQTT broker.")
        socketing.update_client_status(True)
    else:
        print_to_ui(f"Connection to MQTT broker failed: code {rc}")
        socketing.update_client_status(False)


def on_disconnect(client, userdata, rc):
    global connected
    connected = False
    print_to_ui("Disconnected from MQTT broker.")
    socketing.update_client_status(False)


def on_message(client, userdata, message):
    split_topic = message.topic.split("/")
    if len(split_topic) > 1 and split_topic[1] == "status":
        for client_name in status_functions.keys():
            if split_topic[0] == client_name:
                for func in status_functions[client_name]:
                    func(message.payload)


def add_subscription(topic):
    global subscriptions
    subscriptions.append(topic)
    client.subscribe(topic, qos=1)


def subscribe_status(host_name, cb_function):
    global status_functions
    if status_functions.get(host_name) is None:
        status_functions[host_name] = []
    status_functions[host_name].append(cb_function)
    add_subscription(host_name+"/status")


def start_client(url: str, port: int, username: str, pw: str, client_id=CLIENT_ID):
    global CLIENT_ID, client, mqtt_broker_pw, mqtt_broker_url, mqtt_broker_user, mqtt_broker_port
    
    if client_id:
        CLIENT_ID = client_id
    client = mqtt.Client(CLIENT_ID, clean_session=False)
    client.tls_set(certifi.where())
    client.will_set(CLIENT_ID+"/status", "offline", qos=1, retain=True)
    client.reconnect_delay_set(min_delay=1, max_delay=8)
    client.message_retry_set(5)
    
    mqtt_broker_url = url
    mqtt_broker_port = port
    mqtt_broker_user = username
    mqtt_broker_pw = pw
    client.on_connect = on_connect
    client.on_disconnect = on_disconnect
    client.on_message = on_message
    client.username_pw_set(mqtt_broker_user, mqtt_broker_pw)
    client.connect_async(mqtt_broker_url, mqtt_broker_port, keepalive=4)
    client.loop_start()



import flask
import flask_socketio
import logging
import webbrowser
import click

PORT = 5000
webapp = flask.Flask(__name__)
webapp.app_context().push()
socketio = flask_socketio.SocketIO(webapp)
log = logging.getLogger('werkzeug')
log.setLevel(logging.ERROR)


# turn off werkzeug startup printing to the console
def secho(text, file=None, nl=None, err=None, color=None, **styles):
    pass


def echo(text, file=None, nl=None, err=None, color=None, **styles):
    pass


click.echo = echo
click.secho = secho


def port_in_use(port):
    import socket
    with socket.socket(socket.AF_INET, socket.SOCK_STREAM) as s:
        return s.connect_ex(('localhost', port)) == 0


def enough_memory():
    import psutil
    memory = round(psutil.virtual_memory().total / (1024.0 ** 3))
    return memory > 2


def run(port=PORT, nopopup=False):
    if (not nopopup) and enough_memory():
        webbrowser.open(f"http://localhost:{port}", new=1)
    if not port_in_use(port):
        print(f"Ullr started on localhost:{port}")
        socketio.run(webapp, host="0.0.0.0", port=port)
    else:
        print(f"Ullr already running on localhost:{port}")



import os
from configparser import ConfigParser
import uuid
import time

from .getmac import get_mac_address
from .local_device import LocalDevice
from .remote_device import RemoteDevice
from .device_bus import DeviceBus
from .settings import CONFIG_COMMENTS, CONFIG_DEFAULTS
from . import utils, mqtt_client


class UllrConfiguration(object):

    def __init__(self):
        self.parser = ConfigParser()
        self.bus = DeviceBus()
        self.config_file = utils.get_default_config_file()
        self.save_client_id = True

    def _get_client_id(self):
        try:
            CLIENT_ID = get_mac_address().replace(":", "")
        except:
            time.sleep(1)
            utils.logger.warn("getmac process failed. Falling back on uuid.getnode().")
            CLIENT_ID = hex(uuid.getnode())[2:]
            self.save_client_id = False
        return CLIENT_ID

    def add_devices_from_file(self, file=None):
        """ Attempt to add devices from an arbitrary file. If no file is given
            the default file is used.

        """
        if file is None:
            file = self.config_file
        if self._load_config_file(file):
            try:
                self._connect_to_mqtt_broker()
            except Exception as exc:
                utils.print_to_ui(f"Unable to connect to MQTT broker: {exc}")
            try:
                self._add_devices()
            except Exception as exc:
                utils.print_to_ui(
                    f"Invalid config file format: {exc}", sys=True)

    def _connect_to_mqtt_broker(self):
        section = self.parser["$mqtt"]
        broker_url = section["mqtt_broker_url"]
        broker_port = int(section["mqtt_broker_port"])
        broker_user = section["mqtt_broker_user"]
        broker_pw = section["mqtt_broker_pw"]
        client_id = section.get("client_id")
        if not client_id:
            client_id = self._get_client_id()
        mqtt_client.start_client(broker_url, broker_port, broker_user, broker_pw, client_id)

    def _add_devices(self):
        devices = self.parser.sections()
        # make sure there is actually something in config
        if len(devices) > 0:
            utils.print_to_ui(f"Loading devices from config...", sys=True)
            for d in devices:
                name = d
                if name[0] != "$":
                    translation = [False, None, None, 0]
                    location = self.parser[d]["location"]
                    dev_type = self.parser[d]["type"]
                    if self.parser[d]["mute"].upper().strip() == "TRUE":
                        mute = True
                    else:
                        mute = False
                    if self.parser[d]["published"].upper().strip() == "TRUE":
                        published = True
                    else:
                        published = False
                    if self.parser[d]["accepts_incoming"].upper().strip() == "FALSE":
                        accepts_incoming = False
                    else:
                        accepts_incoming = True
                    if self.parser[d]["translated"].upper().strip() == "TRUE":
                        translation[0] = True
                        translation[1] = self.parser[d]["translated_from"]
                        translation[2] = self.parser[d]["translated_to"]
                        translation[3] = int(self.parser[d]["channel_shift"])
                    if location == "local":
                        port = self.parser[d]["port"]
                        baudrate = int(self.parser[d]["baud"])
                        try:
                            dev = LocalDevice(
                                port=port,
                                mode=dev_type,
                                name=name,
                                mute=mute,
                                accepts_incoming=accepts_incoming,
                                baudrate=baudrate,
                                published=published,
                                translation=translation)
                            self.bus.add_device(dev)
                            utils.print_to_ui(
                                f"Added {location} {dev_type} device '{name}' from config.", sys=True)
                        except Exception as e:
                            utils.print_to_ui(
                                f"Failed to add device '{name}' from default config: {e}", sys=True)
                    elif location == "remote":
                        # backwards compatibility for previous versions that used Dweet backend with thing names
                        if self.parser[d].get("thing_name"):
                            topic_name = self.parser[d].get("thing_name")
                        else:
                            topic_name = self.parser[d].get("topic_name") 

                        on_time_max = int(self.parser[d].get("on_time_max"))   
                        try:
                            dev = RemoteDevice(
                                topic_name=topic_name,
                                mode=dev_type,
                                mute=mute,
                                accepts_incoming=accepts_incoming,
                                name=name,
                                on_time_max=on_time_max,
                                translation=translation)
                            self.bus.add_device(dev)
                            utils.print_to_ui(
                                f"Added {location} {dev_type} device '{name}' from config.", sys=True)
                        except Exception as e:
                            utils.print_to_ui(
                                f"Failed to add device '{name}' from default config: {e}", sys=True)
                    else:
                        utils.print_to_ui(f"Failed to add device '{name}' from default config: "
                                        f"invalid location '{location}'", sys=True)

    def save_current_to_file(self):
        """ Save the current device bus to default config file

        """
        self.parser = ConfigParser()  # clear any settings in the parser
        self._add_defaults_to_parser()

        def add_device_to_config(dev):
            if type(dev).__name__ == "LocalDevice":
                self.parser[dev.name]["location"] = "local"
                self.parser[dev.name]["port"] = dev.port_name
                self.parser[dev.name]["baud"] = str(dev.baudrate)
                self.parser[dev.name]["published"] = str(dev.published)
            if type(dev).__name__ == "RemoteDevice":
                self.parser[dev.name]["location"] = "remote"
                self.parser[dev.name]["topic_name"] = dev.topic_name
                self.parser[dev.name]["on_time_max"] = str(dev.on_time_max)
            self.parser[dev.name]["mute"] = str(dev.mute)
            self.parser[dev.name]["accepts_incoming"] = str(dev.accepts_incoming)
            self.parser[dev.name]["translated"] = str(dev.translation[0])
            self.parser[dev.name]["translated_from"] = str(dev.translation[1])
            self.parser[dev.name]["translated_to"] = str(dev.translation[2])
            self.parser[dev.name]["channel_shift"] = str(dev.translation[3])

        # add DCE settings to parser
        for dev in self.bus.dce_devices:
            self.parser.add_section(dev.name)
            self.parser[dev.name]["type"] = "DCE"
            add_device_to_config(dev)

        # add DTE settings to parser
        for dev in self.bus.dte_devices:
            self.parser.add_section(dev.name)
            self.parser[dev.name]["type"] = "DTE"
            add_device_to_config(dev)

        # add MQTT settings to parser
        self.parser["$mqtt"]["mqtt_broker_url"] = mqtt_client.mqtt_broker_url
        self.parser["$mqtt"]["mqtt_broker_port"] = str(mqtt_client.mqtt_broker_port)
        self.parser["$mqtt"]["mqtt_broker_user"] = mqtt_client.mqtt_broker_user
        self.parser["$mqtt"]["mqtt_broker_pw"] = mqtt_client.mqtt_broker_pw
        if self.save_client_id:
            self.parser["$mqtt"]["client_id"] = mqtt_client.CLIENT_ID

        os.makedirs(os.path.dirname(self.config_file), exist_ok=True)

        with open(self.config_file, 'w') as configfile:
            configfile.write(CONFIG_COMMENTS)
        with open(self.config_file, 'a') as configfile:
            self.parser.write(configfile)

        utils.print_to_ui(
            f"Settings saved to config file: {self.config_file}", sys=True)

    def _load_config_file(self, file):
        if not os.path.isabs(file):
            file = os.path.join(os.getcwd(), file)
        if os.path.exists(file):
            utils.print_to_ui(f"Found config file at {file}.", sys=True)
            try:
                self.parser = ConfigParser()
                self.parser.read(file)
                self._add_defaults_to_parser()
                return True
            except Exception as e:
                utils.print_to_ui(f"Failed to read config file: {e}", sys=True)
                return False

        else:
            self._add_defaults_to_parser()
            try:
                self._connect_to_mqtt_broker()
            except Exception as exc:
                utils.print_to_ui(f"Unable to connect to MQTT broker: {exc}")
            utils.print_to_ui(f"Config file {file} not found.", sys=True)
            if file == self.config_file:
                utils.print_to_ui(
                    f"Creating empty default file at: {file}", sys=True)
                self.save_current_to_file()
            return False

    def _add_defaults_to_parser(self):
        for item in CONFIG_DEFAULTS:
            if not self.parser.has_option("DEFAULT", item):
                self.parser["DEFAULT"][item] = CONFIG_DEFAULTS[item]
        if not self.parser.has_section("$mqtt"):
            self.parser.add_section("$mqtt")




# Used to override the default config file location. If set to None the standard will be used,
# which is ~/.config/ullr/config.ini or /etc/ullr/config.ini if superuser
USER_SPECIFIED_DEFAULT_CONFIG_FILE = None

USER_SPECIFIED_LOG_FILE = None

CONFIG_COMMENTS = str("\n; The settings in the [DEFAULT] section reference the values set in 'settings.py' " +
                      "in the package directory." +
                      "\n; They should be here for config file stability." +
                      "\n; If you would like to change these default settings permanently, " +
                      "you should do so in settings.py, not here." +
                      "\n; Make sure you know what you're doing.\n\n")

# defaults when writing new configuration
CONFIG_DEFAULTS = {"type": "",
                   "location": "",
                   "mute": "False",
                   "accepts_incoming": "True",
                   "port": "",
                   "topic_name": "defualt_client/default_device",
                   "baud": "9600",
                   "published": "False",
                   "on_time_max": "0",
                   "translated": "False",
                   "translated_from": "None",
                   "translated_to": "None",
                   "channel_shift": "0",
                   "mqtt_broker_url": "57edaf7763054ccc91c1c8b6e646a155.s1.eu.hivemq.cloud",
                   "mqtt_broker_port": "8883",
                   "mqtt_broker_user": "PiTiming",
                   "mqtt_broker_pw": "Mammoth1"
                   }



import socket
import json
from datetime import date, datetime

from flask import (redirect, render_template, request, Response)

from .. import __version__ as version
from .. import utils
from ..local_device import LocalDevice
from ..remote_device import RemoteDevice
from . import socketing, webapp, socketio
from ..skiracetiming.translate import DECODE, ENCODE
from .. import mqtt_client

current_session = object()


def init(session):
    global current_session
    current_session = session

@webapp.route("/")
def home():
    return render_template(
        "home.html",
        version=version,
        session=current_session,
        ports=utils.get_available_com_ports(),
        hostname=socket.gethostname(),
        host_ip=utils.get_ip(),
        config_file=current_session.config_file.replace("\\", "\\\\"),
        translation_sources=DECODE.keys(),
        translation_destinations=ENCODE.keys(),
        client_id=mqtt_client.CLIENT_ID,
        client_online=mqtt_client.connected,
        broker_url=mqtt_client.mqtt_broker_url,
        broker_port=mqtt_client.mqtt_broker_port,
        broker_user=mqtt_client.mqtt_broker_user,
        broker_pw=mqtt_client.mqtt_broker_pw,
    )


@webapp.route("/add_local", methods=["GET", "POST"])
def add_local():
    if request.method == "POST":
        form = request.form
        mute = False
        incoming = False
        published = False
        if form.get("mute"):
            mute = True
        if form.get("incoming"):
            incoming = True
        if form.get("publish"):
            published = True
        try:
            dev = LocalDevice(
                form["port"],
                form["mode"],
                name=form["name"],
                mute=mute,
                accepts_incoming=incoming,
                baudrate=int(form["baud"]),
                published=published)
            current_session.bus.add_device(dev)
        except Exception as e:
            utils.print_to_ui(f"Failed to add device: {e}")

    return redirect("/")


@webapp.route("/add_remote", methods=["GET", "POST"])
def add_remote():
    if request.method == "POST":
        form = request.form
        mute = False
        incoming = False
        if form.get("mute"):
            mute = True
        if form.get("incoming"):
            incoming = True

        try:
            dev = RemoteDevice(
                form["topic_name"],
                form["mode"],
                name=form["name"],
                accepts_incoming=incoming,
                mute=mute,
                on_time_max=form["on_time_max"]
            )
            current_session.bus.add_device(dev)
        except Exception as e:
            utils.print_to_ui(f"Failed to add device: {e}")

    return redirect("/")


@webapp.route("/remove/<device>", methods=["GET", "POST"])
def remove_device(device):
    current_session.bus.remove_device(device)
    return redirect("/")


@webapp.route("/get_log", methods=["GET", "POST"])
def get_log():
    with open(utils.get_log_file(), "r") as file:
        log = file.read()
    utils.print_to_ui("Served logfile.")
    now = datetime.now().strftime("%Y%m%d_%H%M%S")
    return Response(
        log,
        mimetype="text/plain",
        headers={
            "Content-disposition": f"attachment; filename={now}-ullr-{socket.gethostname()}.log"}
    )


@socketio.on("save_config")
def save_config():
    current_session.save_current_to_file()

@socketio.on("update_mqtt")
def update_mqtt(data):
    data = json.loads(data)
    url, user, pw = "", "", ""
    port = 0
    for item in data:
        if item["name"] == "url":
            url = item["value"]
        elif item["name"] == "port":
            port = int(item["value"])
        elif item["name"] == "user":
            user = item["value"]
        elif item["name"]  == "pw":
            pw = item["value"]
    if mqtt_client.client.is_connected():
        mqtt_client.client.disconnect()
    mqtt_client.start_client(url=url, port=port, username=user, pw=pw)


@socketio.on("update_translation")
def update_translation(id, data):
    data = json.loads(data)
    if data[0]["value"].upper() == "TRUE":
        translation = [
            True,
            data[1]["value"],
            data[2]["value"],
            int(data[3]["value"])
        ]
    else:
        translation = [
            False,
            None,
            None,
            0
        ]
    current_session.bus.update_translation(id, translation)

@socketio.on("update_device")
def update_device(id, data):
    data = json.loads(data)
    d = current_session.bus.find_device(id)
    mute, incoming, publish = False, False, False
    for item in data:
        if item["value"] == "on":
            if item["name"] == "mute":
                mute = True
            elif item["name"] == "incoming":
                incoming = True
            elif item["name"] == "publish":
                publish = True
    d.mute = mute
    d.accepts_incoming = incoming
    if d.type == "serial":
        d.published = publish

@socketio.on("update_ota")
def update_ota(id, data):
    d = current_session.bus.find_device(id)
    data = json.loads(data)
    ota = int(data[0]["value"])
    d.on_time_max = ota
    
@socketio.on("send_late_messages")
def send_late_messages(id, data):
    d = current_session.bus.find_device(id)
    data = json.loads(data)
    index_list = []
    for item in data:
        try:
            index_list.append(int(item["value"]))
        except:
            pass
    d.accept_late_messages(message_index_list=index_list)
    socketio.emit("update_late_messages", id)



import time
from queue import Queue
import threading
import json

import serial
from serial.serialutil import SerialTimeoutException

from .utils import print_to_ui
from . import mqtt_client


class LocalDevice(object):
    """
    A device connected to a serial port on the local machine.
    """

    def __init__(self, port, mode, name="Local Device", mute=False, accepts_incoming=True, baudrate=9600, published=False, translation=[False, None, None, 0]):
        self.sku = id(self)
        if name[0] == "$":
            raise ValueError("Device name cannot start with '$'.")
        self.name = name
        self.type = "serial"
        self.type_color = "red"
        self.port_name = port
        self.mode = mode
        self.accepts_incoming = accepts_incoming
        self.baudrate = baudrate
        self.published = published
        self.published_name = f"{mqtt_client.CLIENT_ID}/{self.name.replace(' ', '_')}"
        if self.published and self.accepts_incoming:
            mqtt_client.client.subscribe(self.published_name+"/from_remote")
            mqtt_client.client.message_callback_add(self.published_name+"/from_remote", self._remote_message)
        self.serial_port = serial.Serial(port=port,
                                         baudrate=baudrate,
                                         write_timeout=3)
        self.buffer = bytearray()
        self._last_message = ''
        self.mute = mute
        self.message_queue = Queue()
        self.translation = translation
        self.exc = False
        self.listening = False
        self.listen_thread = threading.Thread(target=self.listen)
        self.listen_thread.daemon = True
        self.listen_thread.start()

    def write(self, message: bytes):
        """
        Receives a hex string, converts to bytes and writes to the serial port.
        """
        if self.accepts_incoming:
            try:
                self.serial_port.write(message)
                return True
            except SerialTimeoutException:
                return False

    def listen(self):
        """listens to serial port, yields what it hears
        """
        ser = self.serial_port
        self.listening = True

        while self.listening:
            i = self.buffer.find(b"\r")
            while i >= 0:
                if len(self.buffer) > (i+1) and self.buffer[i+1:i+2] == b"\n":
                    i += 1
                message = self.buffer[:i+1]
                self.buffer = self.buffer[i+1:]
                self._last_message = message
                self.message_queue.put(message)
                if self.published:
                    self.publish(message)
                i = self.buffer.find(b"\r")
            time.sleep(.1)
            i = max(1, min(2048, ser.in_waiting))
            data = ser.read(i)
            self.buffer.extend(data)

        ser.close()

        return

    def publish(self, message):
        now = time.time()
        timestamp = now + mqtt_client.time_offset
        payload = json.dumps({"message": message.decode(), "timestamp": timestamp})
        mqtt_client.safe_publish(self.published_name+"/from_device", payload, qos=1)
        print_to_ui("Published to MQTT.")

    def kill_listen_stream(self):
        """
        Stops any threads listening to the serial port.
        """
        self.listening = False

    def get_last_message(self):
        """
        Returns the last message read from the serial port.
        """
        return self._last_message

    def _remote_message(self, client, userdata, message):
        self.write(message.payload)



import datetime

from flask import request

from . import socketio

WEB_CONSOLE_BUFFER = '\r'
TAPES = {}


def timestamp():
    return "[" + datetime.datetime.now().strftime("%Y-%m-%d %H:%M:%S") + "] "


def print_to_web_console(message, endline="\n"):
    global WEB_CONSOLE_BUFFER
    WEB_CONSOLE_BUFFER += str(message) + str(endline)
    socketio.emit("console", WEB_CONSOLE_BUFFER)


def print_tape(target, message):
    message = " ".join(message.split())
    global TAPES
    if TAPES.get(target) is None:
        TAPES[target] = ""
    elif TAPES.get(target) is not str:
        TAPES[target] = str(TAPES[target])
    TAPES[target] += message + "\n"
    payload = {"target": target, "buffer": TAPES[target]}
    socketio.emit("tape_feed", payload)


def load_tapes():
    for tape in TAPES:
        payload = {"target": tape, "buffer": TAPES[tape]}
        socketio.emit("tape_feed", payload)


def update_online_dot(device, online=False):
    payload = {"target": str(device), "online": str(online)}
    socketio.emit("update_online_dot", payload)

def update_late_badge(device, count):
    payload = {"target": str(device), "count": count}
    socketio.emit("update_late_badge", payload)

def update_client_status(online=False):
    payload = {"online": str(online)}
    socketio.emit("update_client_status", payload)

@socketio.on("connect")
def on_connect():
    client_name = request.remote_addr
    load_tapes()
    print_to_web_console(
        f"{timestamp()}Connected to web client: {client_name}.")



# import gevent.monkey; gevent.monkey.patch_all()
import argparse
import sys
import os

from . import __version__ as version
from . import local_device, remote_device, utils, webapp
from .cli import cli
from .session import Session
from .webapp import views

current_session = Session()

# turn off console output if run as windows GUI
if sys.executable.endswith("pythonw.exe"):
    sys.stdout = open(os.devnull, "w")
    sys.stderr = open(os.path.join(os.getenv("TEMP"), "stderr-"+os.path.basename(sys.argv[0])), "w")


def main():
    arg_parser = argparse.ArgumentParser(
        description="An interface for networking R232 devices using dweet.io.")
    options = arg_parser.add_mutually_exclusive_group()
    options.add_argument(
        "--file", type=str, help="Specify config file to use, overriding defaults.")
    options.add_argument("--empty", action='store_true', help="Start Ullr with no devices. "
                                                              "Useful for fixing broken config files "
                                                              "or creating new ones.")
    options.add_argument("--override", metavar=('MODE', 'PORT', 'THING_NAME'), action="store", type=str, nargs=3,
                         help="Setup a basic connection with command line arguments."
                              "\ne.g. --override DCE /dev/ttyUSB0 Ullr_default.")
    arg_parser.add_argument("--nowebui", action="store_true", help="Use Ullr from the command line."
                                                                   "Don't run GUI on webserver.")
    arg_parser.add_argument("--uiport", type=int,
                            default=5000, help="Port for web interface.")
    arg_parser.add_argument("--nopopup", action="store_true",
                            help="Disable automatic opening of browser window to web interface on run.")
    args = arg_parser.parse_args()

    utils.print_to_ui(f"Starting Ullr v{version}...", sys=True)

    if args.override:
        if args.override[0].upper() == "DTE":
            other_mode = "DCE"
        else:
            other_mode = "DTE"
        try:
            local = local_device.LocalDevice(
                args.override[1], args.override[0])
            remote = remote_device.RemoteDevice(args.override[2], other_mode)
            current_session.bus.add_device(local)
            current_session.bus.add_device(remote)
        except Exception as e:
            utils.print_to_ui(f"Override failed: {e}", sys=True)

    elif args.file:
        current_session.add_devices_from_file(args.file)

    elif not args.empty:
        current_session.add_devices_from_file()

    if args.nowebui:
        cli.init(current_session)
        utils.set_ui("cli")
        cli.run()

    else:
        views.init(current_session)
        utils.set_ui("webapp")
        webapp.run(port=args.uiport, nopopup=args.nopopup)


if __name__ == "__main__":
    main()



# All encode functions take a channel and tod (time of day) argument. The
# channel argument is a list. The first item on the list is an integer
# representing the actual channel number. The channel number should be
# normalized, so the first channel is 0, whether or not the timer itself starts
# its channel counts at 0. The second item in the channel list argument is a
# boolean value indicating whether or not the impulse was manual, eg. from a
# pushbutton on the timer. So, a channel argument of [0, True] would compute to
# "c0M" on an Alge Timy, because Timy channels start at 0. It would compute to
# "M1" on a Tag CP540 because CP540 channels start at 1, and "SZ " on an s4
# because s4s are weird. The same logic applies to decoding.

class TranslationError(Exception):
    pass


def _decode_cp540(message):
    chan = message[13:15]
    tod = message[16:29]
    channel = ['', False]
    if chan[0].upper() == "M":
        channel[1] = True
    # First channel of CP540 is 1. This normalizes to first channel of 0
    channel[0] = int(chan[1]) - 1
    return channel, tod


def _encode_cp540(channel, tod):
    channel[0] += 1  # First channel of CP540 is 1, not 0.
    if channel[0] > 16 or channel[0] < 1:
        raise TranslationError(f"Channel {channel[0]} out of range for CP540.")
    if channel[0] > 9:
        chan = f"{channel[0]}"
    elif channel[1]:
        chan = f"M{channel[0]}"
    else:
        chan = f" {channel[0]}"
    msg = f"TN           {chan} {tod}      "
    # calculate checksum
    cs16 = f'{int(sum(bytearray(msg.encode("ascii"))) % 65536):04x}'.upper()
    msg += f"\t{cs16}\r\n"
    return msg


def _decode_timy(message):
    chan = message[6:9]
    tod = message[10:23]
    channel = ["", False]
    if chan[2].upper() == "M":
        channel[1] = True
    channel[0] = int(chan[1])
    return channel, tod


def _encode_timy(channel, tod):
    if channel[0] > 7 or channel[0] < 0:
        raise TranslationError(f"Channel {channel[0]} out of range for Timy.")
    # only keep manual indicator for channels 0 and 1, as only those buttons exist on Timy
    if channel[1] and channel[0]<2:
        chan = f"c{channel[0]}M"
    else:
        chan = f"c{channel[0]} "
    msg = f"      {chan} {tod}       \r"
    return msg


def _decode_s4(message):
    chan = message[5:8]
    tod = message[9:21] + "0"  # add zero to normalize TOD to 1/10,000 format
    channel = ["", False]
    if chan.upper()[:2] == "SZ":
        channel[0] = 0
    else:
        channel[0] = int(chan[1:])
    return channel, tod


def _encode_s4(channel, tod):
    if channel[0] > 17 or channel[0] < 0:
        raise TranslationError(f"Channel {channel[0]} out of range for S4.")
    if channel[0] == 0:
        chan = "SZ "
    else:
        chan = f"K{channel[0]:02d}"
    tod = str(tod)[:12]  # remove last digit as S4 only goes to 1/1,000s
    msg = f"0000 {chan} {tod}\r"
    return msg

# Racetime2 format:
# QQQQSSSSRRRCCCLLLIOPTTTTTTTTT<CR>
# Where Q = Sequence #
#       S = Start #
#       R = Run #
#       C = Physical channel (000 start, 255 finish, 254 "aux", 1-253 "lap")
#       L = Logical channel, same as above
#       I = "info", 0 for "time event"
#       O = Origin, "M" for manual or physical input, "R" for radio
#       P = Sign, 1 for negative, space otherwise
#       T = Time with no ":", HHMMSSttt
# Example: "000100010012552550M 123456789\r" Finish at 12:34:56.789, either manually or from cable


def _decode_racetime2(message):
    chan = int(message[14:17])
    channel = ["", False]
    # try to normalize channel numbers so they can be usable by other formats.
    # finish goes from 255>1, aux from 254>2, radio channels start at 3 instead of 1.
    if chan == 0:
        channel[0] = 0
    elif chan == 255:
        channel[0] = 1
    elif chan == 254:
        channel[0] = 2
    else:
        channel[0] = chan + 2
    tod = f"{message[20:22]}:{message[22:24]}:{message[24:26]}.{message[26:29]}"
    return channel, tod


def _encode_racetime2(channel, tod):
    # try to map normalized channel numbers to the physical channels of racetime2
    if channel[0] > 255 or channel[0] < 0:
        raise TranslationError(
            f"Channel {channel[0]} out of range for RaceTime2.")
    if channel[0] == 0:
        chan = "000"
    elif channel[0] == 1:
        chan = 255
    elif channel[0] == 2:
        chan = 254
    else:
        chan = f"{(channel[0]-2):03d}"
    tod = tod[:12].replace(":", "").replace(
        ".", "")  # remove last digit and punctuation
    msg = f"00010001001{chan}{chan}0M {tod}\r"
    return msg


ENCODE = {
    "cp540": _encode_cp540,
    "timy": _encode_timy,
    "s4": _encode_s4,
    "racetime2": _encode_racetime2
}

DECODE = {
    "cp540": _decode_cp540,
    "timy": _decode_timy,
    "s4": _decode_s4,
    "racetime2": _decode_racetime2
}


def translate(message, source, destination, shift=0):
    channel, tod = DECODE[source](message)
    channel[0] += shift
    translation = ENCODE[destination](channel, tod)
    return translation



import time

from colorama import init as colorama_init
from termcolor import colored

from .. import __version__ as version
from ..local_device import LocalDevice
from ..remote_device import RemoteDevice
from . import interface

current_session = object()
colorama_init()


def init(session):
    global current_session
    current_session = session


def add_device():
    name = interface.s_input("\nDevice Name: ")
    location = interface.s_input("Location (1.local 2.remote): ")
    mode = interface.s_input("Type (DCE/DTE): ").upper()
    mute = interface.s_input("Mute device? (y/n): ")
    if mute.upper().strip() == "Y" or mute.upper().strip() == "YES":
        mute = True
    else:
        mute = False
    d = None

    if location == "1":
        port = interface.s_input("Port: ")
        baudrate = interface.s_input("Baudrate: ")
        if baudrate == "":
            baudrate = 9600
        try:
            d = LocalDevice(
                port=port,
                mode=mode,
                name=name,
                mute=mute,
                baudrate=int(baudrate)
            )
        except Exception as e:
            interface.s_print(e)

    elif location == "2":
        thing_id = interface.s_input("Thing ID: ")
        key = interface.s_input("Thing Key: ")
        if key == "None" or key == "":
            key = None
        try:
            d = RemoteDevice(thing_id, mode, key, name, mute)
        except Exception as e:
            interface.s_print(e)
    else:
        interface.s_print("Invalid input")
        return
    if d:
        current_session.bus.add_device(d)
    return


def remove_device():
    current_session.bus.print_status()
    device = interface.s_input("\nDevice to remove: ")
    current_session.bus.remove_device(device)
    return


def process_input(cmd, ):
    if cmd == "info":
        return interface.s_print(interface.get_devices_table(current_session))
    if cmd == "threads":
        return current_session.bus.print_threads()
    if cmd == "add":
        return add_device()
    if cmd == "remove":
        return remove_device()
    if cmd == "save":
        return current_session.save_current_to_file()
    else:
        # print command help
        interface.s_print("\tType 'info' to display session info.\n"
                          "\tType 'add' to add a device.\n"
                          "\tType 'remove' to remove a device.\n"
                          "\tType 'save' to save the current configuration as default."
                          )
        return


def idle():
    while True:
        time.sleep(1)


def menu():
    while True:
        cmd = ''
        try:
            time.sleep(.0001)
            cmd = interface.s_input(
                "\nType 'exit' to exit or ENTER for help.\n")
        except EOFError:  # if ran as a daemon, make sure we don't reach EOF prematurely
            idle()
        if cmd == 'exit':
            break
        process_input(cmd)


def run():
    interface.s_print("\t\t*************************************************\n"
                      "\t\t**                   " + colored("dweet", "cyan") + "2" +
                      colored("ser", "red") + "                 **\n"
                      f"\t\t**                    v{version}                   **\n"
                      "\t\t*************************************************\n")
    menu()



from threading import Lock

from termcolor import colored
from colorama import init as colorama_init, Fore, Style

colorama_init()
s_print_lock = Lock()


def s_print(*a, **b):
    """Thread safe print function"""
    with s_print_lock:
        print(*a, **b)


def s_input(*a):
    """Thread safe input function"""
    with s_print_lock:
        print(*a, end='')
    return input('')


def get_devices_table(session):
    """
    Returns a list of devices in table form.
    """
    def get_table(dev_list):
        cols = ["#".ljust(3),
                "Name".ljust(16),
                "Type".ljust(10),
                "Port".ljust(15),
                "ThingName".ljust(20),
                "Locked".ljust(10),
                "Muted".ljust(10)
                ]
        buf = ''
        header = ''
        for col in cols:
            header = header + f"{col}  "
        buf += (f"\t{Fore.LIGHTWHITE_EX}{header}{Style.RESET_ALL}\n")

        for i in range(0, len(dev_list)):
            num = str(i + 1)
            d = dev_list[i]
            if type(d).__name__ == "LocalDevice":
                buf += (f"\t"
                        f"{num.ljust(3)}  "
                        f"{d.name.ljust(16)}  "
                        f"{colored(d.type.ljust(10), d.type_color)}  "
                        f"{d.port_name.ljust(15)}  "
                        f"{''.ljust(20)}  "
                        f"{''.ljust(10)}  "
                        f"{str(d.mute).ljust(10)}  "
                        )
            if type(d).__name__ == "RemoteDevice":
                buf += (f"\t"
                        f"{num.ljust(3)}  "
                        f"{d.name.ljust(16)}  "
                        f"{colored(d.type.ljust(10), d.type_color)}  "
                        f"{''.ljust(15)}  "
                        f"{d.thing_id.ljust(20)}  "
                        f"{str(d.locked).ljust(10)}  "
                        f"{str(d.mute).ljust(10)}  "
                        )
        buf += "\n"
        return buf
    buffer = ''
    buffer += "\nDCE Devices\n"
    buffer += get_table(session.bus.dce_devices)
    buffer += "\nDTE Devices\n"
    buffer += get_table(session.bus.dte_devices)
    return buffer



import threading
import time
from serial import SerialException

from colorama import Fore, Style
from colorama import init as colorama_init
from termcolor import colored
from urllib3.exceptions import ProtocolError

from .utils import internet_connection, print_to_ui, logger
from .webapp.socketing import print_tape, TAPES
from . import skiracetiming

colorama_init()

class DeviceBus(object):
    """
    Holds all the devices in a connection and facilitates communication between them. All DCE devices
    write to all DTE devices and vice versa.
    """

    def __init__(self):
        self.dce_devices = []
        self.dte_devices = []
        self.listen_threads = {}


    def add_device(self, device):
        """
        Adds the given device to the connection, and starts a listening thread if the device is not muted.
        """
        if device.mode == "DTE":
            self.dte_devices.append(device)
        if device.mode == "DCE":
            self.dce_devices.append(device)
        if not device.mute:
            self.listen_threads[device.name] = threading.Thread(
                target=self._listen_stream, args=[device])
            self.listen_threads[device.name].daemon = True
            self.listen_threads[device.name].start()
        print_to_ui(f"Added device {device.name}.")
        return True

    def find_device(self, key):
        for d in self.dce_devices:
            if d.name == key:
                return d
            elif str(d.sku) == key:
                return d
        for d in self.dte_devices:
            if d.name == key:
                return d
            elif str(d.sku) == key:
                return d
        return None

    def remove_device(self, name_or_sku):
        """
        Takes a device name, and removes that device if the name is found
        """
        device = self.find_device(name_or_sku)

        if device:
            device.kill_listen_stream()
            if device.mode.lower() == "dce":
                self.dce_devices.remove(device)
            elif device.mode.lower() == "dte":
                self.dte_devices.remove(device)
            TAPES.pop(device.sku, "")
            print_to_ui(f"Device '{device.name}' removed.")

        else:
            print_to_ui(f"Device {name_or_sku} not found.")

    def update_translation(self, name_or_sku, translation_list):
        device = self.find_device(name_or_sku)
        if device:
            try:
                device.translation = translation_list
                print_to_ui(
                    f"Updated translation for {device.name}: {translation_list}")
            except Exception as e:
                print_to_ui(
                    f"Could not update translation for {device.name}: {e}")
        else:
            print_to_ui(f"Device {name_or_sku} not found.")

    def print_threads(self):
        print_to_ui(self.listen_threads)

    def _listen_stream(self, device):
        """
        Listens for messages from the given device, then writes to all appropriate devices on bus.
        """
        print_to_ui(f"Listen stream started for {device.name}.")

        while device.listening:
            try:
                message = device.message_queue.get()
                try:
                    message_decoded = message.decode('latin-1').rstrip().replace('\r', '')
                except ValueError as e:
                    logger.warn(f"Received non-bytes message: {message}")
                    print_to_ui(f"{message}: {e}")
                    message_decoded = "<indecipherable>"

                print_tape(device.sku, message_decoded)
                print_to_ui(f"Received {colored(device.type, device.type_color)} message from {device.name}:"
                            f" {Fore.LIGHTWHITE_EX}{message_decoded}{Style.RESET_ALL}")

                if device.translation[0]:
                    try:
                        message_decoded = skiracetiming.translate(message_decoded, device.translation[1],
                                                                    device.translation[2], device.translation[3])
                        message = message_decoded.encode()
                        message_decoded = message_decoded.rstrip().replace('\r', '')
                        print_to_ui(f"Translated from {device.translation[1]} to {device.translation[2]} "
                                    f"with channel shift {device.translation[3]}.")

                    except Exception as e:
                        print_to_ui(f"Translation failed: {e}")

                def write_to_device_list(list):
                    for d in list:
                        if d.accepts_incoming:
                            if d.write(message):
                                print_to_ui(
                                    f"{colored(d.type.capitalize(), d.type_color)} message sent to {d.name}: "
                                    f"{message_decoded}")
                            else:
                                print_to_ui(f"Writing to {d.name} failed.")

                if device.mode == "DTE":
                    # Messages from DTE devices get sent to all DCE devices.
                    write_to_device_list(self.dce_devices)

                elif device.mode == "DCE":
                    # Messages from DCE devices get sent to all DTE devices.
                    write_to_device_list(self.dte_devices)

                else:
                    print_to_ui("Mode not found")
                    
            except SerialException as e:
                print_to_ui(f"{e}")
                print_to_ui(f"{device.name} unplugged. Removing device.")
                self.remove_device(device.name)

        return True





try:
    import setuptools

except ImportError:
    from distutils.core import setup
    from io import open

from ullr import __version__ as version

with open('README.md', 'r', encoding='utf-8') as fh:
    long_description = fh.read()

setuptools.setup(
    name='Ullr',
    version=version,
    author='Zach Henry',
    author_email='zhenry9@gmail.com',
    description='A serial <-> MQTT interface for sports timing.',
    long_description=long_description,
    long_description_content_type='text/markdown',
    url='https://ullr.rtfd.io',
    # download_url=f'https://github.com/zhenry9/ullr/archive/{version}.tar.gz',
    packages=setuptools.find_packages(exclude=('tests',)),
    include_package_data=True,
    classifiers=[
        'Programming Language :: Python :: 3',
        'Development Status :: 3 - Alpha',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: GNU General Public License (GPL)',
        'Natural Language :: English',
    ],
    install_requires=[
        "astroid~=2.8.5",
        "bidict~=0.21.4",
        "certifi~=2021.10.8",
        "chardet~=4.0.0",
        "charset-normalizer~=2.0.7",
        "click~=8.0.3",
        "colorama~=0.4.4",
        "dnspython~=2.1.0",
        "docopt~=0.6.2",
        "Flask~=2.0.2",
        "Flask-SocketIO~=5.1.1",
        "idna~=3.3",
        "importlib-metadata~=4.8.2",
        "isort~=5.10.1",
        "itsdangerous~=2.0.1",
        "Jinja2~=3.0.3",
        "lazy-object-proxy~=1.6.0",
        "MarkupSafe~=2.0.1",
        "mccabe~=0.6.1",
        "ntplib~=0.4.0",
        "paho-mqtt~=1.6.1",
        "platformdirs~=2.4.0",
        "psutil~=5.8.0",
        "pylint~=2.11.1",
        "pyserial~=3.5",
        "python-engineio~=4.3.0",
        "python-socketio~=5.5.0",
        "requests~=2.26.0",
        "six~=1.16.0",
        "termcolor~=1.1.0",
        "toml~=0.10.2",
        "typed-ast~=1.5.0",
        "typing-extensions~=4.0.0",
        "urllib3~=1.26.7",
        "Werkzeug~=2.0.2",
        "wrapt~=1.13.3",
        "yarg~=0.1.9",
        "zipp~=3.6.0",
        "zope.event~=4.5.0",
        "zope.interface~=5.4.0"
    ],
    python_requires='>=3.6',
    entry_points={
        'console_scripts': [
            'ullr = ullr.__main__:main',
        ]
    },
    zip_safe=False,
)
